/*
  findprime.c

  Uses wheel factorization to generate a prime in a certain range.
  See http://www.utm.edu/research/primes/glossary/WheelFactorization.html

  Part of my implementation of the RSA scheme
*/

#include "findprime.h"

/* Seeds the random number generator from /dev/random */
void seedRandom (void)
{
  /* An array for the random function so we can get strong random numbers */
  /* Declared static so it is maintained between function calls */
  static char randomState[256];
  /* We need to read in data from /dev/random to seed the generator well */
  FILE * input;
  
  /* Initialize the array so it can be seeded */
  initstate (0, &randomState, 256);

  /* Seed the generator with /dev/random data */
  if ( (input = fopen ("/dev/random", "r")) == NULL )
    {
      fprintf (stderr, "Cannot open /dev/random\n");
      exit (1);
    }
  if ( fread (&randomState, sizeof (char), 256, input) != 256)
    {
      fprintf (stderr, "Error reading /dev/random\n");
      exit (2);
    }
  fclose (input);

  /* Finally set the state */
  setstate (&randomState);

  /* Test the generator */
  /*
  for (testProduct = 0; testProduct < 10; ++ testProduct)
    {
      printf ("%u\n", random () );
    }
  */

  return;
}
/*
   Randomly finds a prime number between lower and upper using wheel
   factorization
   (http://www.utm.edu/research/primes/glossary/WheelFactorization.html) to
   test for primality. This could be *much* faster with advanced primality
   tests and expecially with probabilistic primality testing, which any
   serious RSA implementation should use. The lower bound is included, the
   upper bound is excluded.

   Return values:
   x	Some prime number
*/
unsigned long int findPrime (unsigned long int lower, unsigned long int upper)
{
  /* A bunch of arrays that we will use to test primality */
  unsigned long int test [] = { 2, 3, 5, 7 };
  unsigned long int divisors [] = { 1, 11, 13, 17, 19, 23, 29, 31, 37, 41,
				    43, 47, 53, 59, 61, 67, 71, 73, 79, 83,
				    89, 97, 101, 103, 107, 109, 113, 121, 127, 131,
				    137, 139, 143, 149, 151, 157, 163, 167, 169, 173,
				    179, 181, 187, 191, 193, 197, 199, 209 };
  unsigned long int numDivisors = 48, numTest = 4, newPrime = 0, testProduct = 2 * 3 * 5 * 7;
  unsigned long int root = 0, multiple = 0, counter = 0, currentPrime = 0;

  //printf ("Generating possible primes between %lu and %lu...\n", lower, upper);

  /* Now we generate a random number in the specified range and do
     wheel factorization to verify primality until we find a prime */
  /* Subtract because we are going to add 2 */
  do
    {
      newPrime = random () % (upper - lower) + lower - 2;	  
    } 
  while (newPrime % 2 == 0);
  
  currentPrime = 0;
  
  while (! currentPrime)
    {
      /* Test the next odd number */
      newPrime += 2;      
      /* Handle the passed the top problem and the overflow the integer problem */
      if (newPrime > upper)
	{
	  newPrime -= lower;
	}
      if (newPrime < lower)
	{
	  newPrime += lower;
	}
      currentPrime = newPrime;
      //printf ("%lu...", currentPrime);

      /* Test it with the test divisors first, skipping 2 */
      for (counter = 1; counter < numTest && currentPrime; ++ counter)
	{
	  if (currentPrime % test [counter] == 0)
	    {
    	      //printf (" factor = %lu\n", test [counter]);
              currentPrime = 0;
	    }
	}

      if (currentPrime)
	{
	  root = ceil ( sqrt (currentPrime) );

	  /* We stop testing divisors when we have passed the square root */
	  /* Skip 1 when we start testing */
	  for (counter = 1, multiple = 0; currentPrime && divisors [counter] + multiple <= root; ++ counter)
	    {
	      if (currentPrime % (divisors [counter] + multiple) == 0)
		{
		  //printf (" factor = %lu\n", divisors [counter] + multiple);	       
		  currentPrime = 0;
		}

	      /* Increase the multiple when we need to */
	      if (counter == numDivisors)
		{
		  counter = 0;
		  multiple += testProduct;
		}
	    }
	}
      
    } 

  //printf (" is prime.\n");
  return currentPrime;
}
